@@ -2,7 +2,7 @@
 import json
 import re
 import functools
-from typing import List, Dict, Tuple, Optional, TypeVar, Iterable
+from typing import List, Dict, Set, Tuple, Optional, TypeVar, Iterable
 from spacy.tokens import Doc, Token, Span, DocBin  # type: ignore
 import numpy as np
 
@@ -31,6 +31,7 @@ def is_likely_proper(tok: Token, min_rank=200) -> bool:
     # If there is no lemma, but the token is in uppercase, return true as well
     elif tok.lemma_=="" and tok.is_upper:
         return True
+
     
     # We do not consider the 200 most common words as proper name
     elif (tok.lemma_.islower() and tok.lemma in tok.vocab.strings
@@ -58,8 +59,9 @@ def is_likely_proper(tok: Token, min_rank=200) -> bool:
     elif tok.pos_ == "PROPN":
         return True
 
-    # If the token is in lowercase but is a quite rare token
-    elif len(tok) > 3 and tok.is_lower and len(tok.vocab.vectors) > 0 and tok.is_oov:
+    # If the token is a quite rare token
+    elif (len(tok) > 3 and (tok.is_lower or tok.is_upper)
+          and len(tok.vocab.vectors) > 0 and tok.is_oov):
         return True
 
     return False
@@ -437,8 +439,11 @@ def get_subsequences(sequence: List[T]) -> List[List[T]]:
     return subsequences
 
 
-def spans_to_array(doc: Doc, labels: List[str],
-                   sources: List[str] = None) -> np.ndarray:
+def spans_to_array(
+    doc: Doc,
+    labels: List[str],
+    sources: Optional[List[str]] = None,
+) -> np.ndarray:
     """Convert the annotations of a spacy document into a 2D array.
     Each row corresponds to a token, and each column to a labelling
     source. In other words, the value at (i,j) represents the prediction
@@ -452,45 +457,144 @@ def spans_to_array(doc: Doc, labels: List[str],
     NB: we assume the labels use either IO/BIO/BILUO, and that the
     O label is at position 0.
     """
+    label2idx, prefixes, labels_without_prefix, = _index_labels(
+        original_labels=labels,
+        strip_prefixes=False,
+    )
+    return _spans_to_array(
+        doc, 
+        sources,
+        label2idx,
+        labels_without_prefix,
+        prefixes
+    )
 
-    # Creating some helper dictionaries
-    label_indices = {}
+
+def _index_labels(
+    original_labels: List[str],
+    strip_prefixes: bool = False
+) -> Tuple[Dict[str, int], Set[str], Set[str]]:
+    """ Normalize and index a list of labels to:
+    1. Generate a mapping from labels to indices
+    2. Identify label prefixes (e.g., I, B, L, etc.)
+    3. Identify list of labels without prefixes
+
+    If `strip_prefixes` is True, we normalize IO/BIO/BILUO labels
+    (e.g., B-PERSON and I-PERSON will be normalized PERSON). 
+    
+    If `strip_prefixes` is False, we assume the labels
+    use IO/BIO/BILUO formats and mantain different labels/indices for labels
+    like B-PERSON and I-PERSON.
+
+    NB: We assume that the first label in labels is 'O' for the null token.
+    """
+    labels = []
     prefixes = set()
     labels_without_prefix = set()
-    for i, label in enumerate(labels):
-        label_indices[label] = i
-        if "-" in label:
-            prefix, label = label.split("-", 1)
-            prefixes.add(prefix)
-            labels_without_prefix.add(label)
 
+    for original_label in original_labels:
+        if "-" in original_label:
+            # Normalize B-PER and I-PER to PER
+            prefix, normalized_label = original_label.split("-", 1)
+            prefixes.add(prefix)
+        else:
+            # No normalization required 
+            normalized_label = original_label
+
+        # Track labels without prefixes
+        labels_without_prefix.add(normalized_label)
+
+        if strip_prefixes:
+            # Use normalized label for token labeling
+            if normalized_label not in labels:
+                labels.append(normalized_label)
+        else:
+            # Use original label for token labeling
+            if original_label not in labels:
+                labels.append(original_label)
+
+    # Generate mapping of labels to label indices
+    label2idx = {label: i for i, label in enumerate(labels)}
+
+    return label2idx, prefixes, labels_without_prefix
+
+
+def _spans_to_array(
+    doc: Doc,
+    sources: List[str],
+    label2idx: Dict[str, int],
+    labels_without_prefix: Set[str],
+    prefixes: Optional[Set[str]] = None,
+    warn_missing_labels: bool = False
+) -> np.ndarray:
+    """Convert the annotations of a spacy document into a 2D array.
+    Each row corresponds to a token, and each column to a labelling
+    source. In other words, the value at (i,j) represents the prediction
+    of source j for token i. This prediction is expressed as the
+    index of the label in the labels.
+
+    NB:
+        - Sources should be a list of labelling sources. If empty, all sources
+            are employed.
+        - If `prefixes` are provided (e.g., [I, B, L]), it is assumed that the 
+            labels in `label2idx` contain the prefixes (e.g., I-PERSON,
+            B-PERSON).
+        - If `prefixes` are not provided, it is assumed that the labels in 
+            `label2idx` do not contain prefixes (e.g, PERSON). 
+        - We also assume the O is label is at position 0.
+    """
     if sources is None:
         sources = list(doc.spans.keys())
+    
+    if warn_missing_labels:
+        missing_labels = set()
 
     # Creating the numpy array itself
     data = np.zeros((len(doc), len(sources)), dtype=np.int16)
 
     for source_index, source in enumerate(sources):
         for span in doc.spans.get(source, []):
-
             if span.label_ not in labels_without_prefix:
+                if warn_missing_labels:
+                    missing_labels.add(span.label_)
                 continue
 
-            # If the span is a single token, we can use U
-            if "U" in prefixes and len(span) == 1:
-                data[span.start, source_index] = label_indices["U-%s" % span.label_]
-                continue
+            if prefixes is None:
+                # Do not use prefix labels (e.g., use PER instead of 
+                # B-PER, I-PER, etc.) 
+                data[span.start:span.end, source_index] = label2idx[
+                    span.label_
+                ]
+            else:
+                # If the span is a single token, we can use U
+                if "U" in prefixes and len(span) == 1:
+                    data[span.start, source_index] = label2idx[
+                        "U-%s" % span.label_
+                    ]
+                    continue
 
-            # Otherwise, we use B, I and L
-            if "B" in prefixes:
-                data[span.start, source_index] = label_indices["B-%s" % span.label_]
-            if "I" in prefixes:
-                start_i = (span.start+1) if "B" in prefixes else span.start
-                end_i = (span.end-1) if "L" in prefixes else span.end
-                data[start_i:end_i, source_index] = label_indices["I-%s" % span.label_]
-            if "L" in prefixes:
-                data[span.end-1, source_index] = label_indices["L-%s" % span.label_]
+                # Otherwise, we use B, I and L
+                if "B" in prefixes:
+                    data[span.start, source_index] = label2idx[
+                        "B-%s" % span.label_
+                    ]
+                if "I" in prefixes:
+                    start_i = (span.start+1) if "B" in prefixes else span.start
+                    end_i = (span.end-1) if "L" in prefixes else span.end
+                    data[start_i:end_i, source_index] = label2idx[
+                        "I-%s" % span.label_
+                    ]
+                if "L" in prefixes:
+                    data[span.end-1, source_index] = label2idx[
+                        "L-%s" % span.label_
+                    ]
 
+    if warn_missing_labels:
+        print(
+            "WARNING: \
+            Span labels were found in the dataset that were not provided \
+            in `labels_without_prefices`: {}".format(missing_labels)
+        )
     return data
 
 
@@ -542,8 +646,8 @@ def token_array_to_spans(agg_array: np.ndarray,
     return spans
 
 
-def token_array_to_probs(agg_array: np.ndarray,
-                         prefix_labels: List[str]) -> Dict[int, Dict[str, float]]:
+def token_array_to_probs(agg_array: np.ndarray, prefix_labels: List[str],
+                         min_threshold=0.1) -> Dict[int, Dict[str, float]]:
     """Given a 2D array containing, for each token, the probabilities for a
     each possible output label in prefix form (B-PERSON, I-ORG, etc.), returns
     a dictionary of dictionaries mapping token indices to probability distributions
@@ -557,10 +661,13 @@ def token_array_to_probs(agg_array: np.ndarray,
     # We only look at labels beyond "O", and with non-zero probability
     row_indices, col_indices = np.nonzero(agg_array[:, 1:])
     for i, j in zip(row_indices, col_indices):
-        if i not in token_probs:
-            token_probs[i] = {prefix_labels[j+1]: agg_array[i, j+1]}
+        prob = agg_array[i, j+1]
+        if prob < min_threshold:
+            continue
+        elif i not in token_probs:
+            token_probs[i] = {prefix_labels[j+1]: prob} #type: ignore
         else:
-            token_probs[i][prefix_labels[j+1]] = agg_array[i, j+1]
+            token_probs[i][prefix_labels[j+1]] = prob #type: ignore
 
     return token_probs
 
@@ -612,11 +719,14 @@ def is_valid_transition(prefix_label1, prefix_label2, encoding="BIO"):
 # Visualisation
 ############################################
 
-def display_entities(doc: Doc, layer=None, add_tooltip=True):
+def display_entities(doc: Doc, layer=None, add_tooltip=False):
     """Display the entities annotated in a spacy document, based on the
     provided annotation layer(s). If layer is None, the method displays
     the entities from Spacy. 
     This method will only work in a Jupyter Notebook or similar. 
+    If add_tooltip is set to True, the visualisation also adds tooltips to show
+    the predictions of each labelling functions for a given token. This functionality
+    only works with Jupyter Lab (not Jupyter Notebook).
     """
     import spacy.displacy
     import IPython.core.display
@@ -630,7 +740,7 @@ def display_entities(doc: Doc, layer=None, add_tooltip=True):
                               if re.match(layer.replace("*", ".*?")+"$", l)]
             spans = get_spans(doc, matched_layers)
         else:
-            spans = doc.spans[layer]
+            spans = get_spans(doc, [layer])
     else:
         raise RuntimeError("Layer type not accepted")
 
@@ -733,4 +843,4 @@ def _enrich_with_tooltip(doc: Doc, html: str, sources: List[str]):
 </style>
 """ + "".join(new_fragments)
 
-    return new_html
+    return new_html
\ No newline at end of file
